@@ -33,7 +33,7 @@
  * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
  * THE POSSIBILITY OF SUCH DAMAGES.
  *
- * $Id: ieee80211_wireless.c,v 1.23 2005/07/19 18:19:19 br1 Exp $
+ * $Id: ieee80211_wireless.c 1187 2005-10-19 15:34:16Z br1 $
  */
 #ifndef EXPORT_SYMTAB
 #define	EXPORT_SYMTAB
@@ -617,9 +617,10 @@ ieee80211_ioctl_siwap(struct ieee80211com *ic,
 {
 	static const u_int8_t zero_bssid[IEEE80211_ADDR_LEN];
 
-	/* NB: should only be set when in STA mode */
+	/* NB: should not be set when in HOSTAP mode */
 	if (ic->ic_opmode != IEEE80211_M_STA &&
-	    ic->ic_opmode != IEEE80211_M_AHDEMO)
+	    ic->ic_opmode != IEEE80211_M_AHDEMO &&
+	    ic->ic_opmode != IEEE80211_M_IBSS)
 		return -EINVAL;
 	IEEE80211_ADDR_COPY(ic->ic_des_bssid, &ap_addr->sa_data);
 	/* looks like a zero address disables */
@@ -1607,6 +1608,9 @@ ieee80211_ioctl_setparam(struct ieee80211com *ic, struct iw_request_info *info,
 		if (ic->ic_curmode == IEEE80211_MODE_11G)
 			retv = ENETRESET;
 		break;
+	case IEEE80211_PARAM_WDSONLY:
+		ic->ic_wdsonly = value;
+		break;
 	case IEEE80211_PARAM_RESET:
 		ic->ic_init(ic->ic_dev);
 		break;
@@ -1744,6 +1748,9 @@ ieee80211_ioctl_getparam(struct ieee80211com *ic, struct iw_request_info *info,
 	case IEEE80211_PARAM_PUREG:
 		param[0] = (ic->ic_flags & IEEE80211_F_PUREG) != 0;
 		break;
+	case IEEE80211_PARAM_WDSONLY:
+		param[0] = ic->ic_wdsonly;
+		break;
 	default:
 		return -EOPNOTSUPP;
 	}
@@ -1829,6 +1836,7 @@ ieee80211_ioctl_setkey(struct ieee80211com *ic, struct iw_request_info *info,
 			return -EINVAL;
 		wk = &ic->ic_nw_keys[kid];
 		ni = NULL;
+		ik->ik_flags |= IEEE80211_KEY_GROUP;
 	}
 	error = 0;
 	ieee80211_key_update_begin(ic);
@@ -2047,9 +2055,9 @@ ieee80211_ioctl_setmlme(struct ieee80211com *ic, struct iw_request_info *info,
 			return -ENOENT;
 		}
 		if (mlme->im_op == IEEE80211_MLME_AUTHORIZE)
-			ieee80211_node_authorize(ic, ni);
+			ieee80211_node_authorize(ni);
 		else
-			ieee80211_node_unauthorize(ic, ni);
+			ieee80211_node_unauthorize(ni);
 		ieee80211_free_node(ni);
 		break;
 	case IEEE80211_MLME_CLEAR_STATS:
@@ -2230,6 +2238,10 @@ static const struct iw_priv_args ieee80211_priv_args[] = {
 	  IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,"delmac" },
 	{ IEEE80211_IOCTL_CHANLIST,
 	  IW_PRIV_TYPE_CHANLIST | IW_PRIV_SIZE_FIXED, 0,"chanlist" },
+	{ IEEE80211_IOCTL_WDSADD,
+	  IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,"wdsadd" },
+	{ IEEE80211_IOCTL_WDSDEL,
+	  IW_PRIV_TYPE_ADDR | IW_PRIV_SIZE_FIXED | 1, 0,"wdsdel" },
 #if WIRELESS_EXT >= 12
 	{ IEEE80211_IOCTL_SETPARAM,
 	  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 2, 0, "setparam" },
@@ -2353,6 +2365,10 @@ static const struct iw_priv_args ieee80211_priv_args[] = {
 	  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "pureg" },
 	{ IEEE80211_PARAM_PUREG,
 	  0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "get_pureg" },
+	{ IEEE80211_PARAM_WDSONLY,
+	  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "wdsonly" },
+	{ IEEE80211_PARAM_WDSONLY,
+	  0, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, "get_wdsonly" },
 	{ IEEE80211_PARAM_RESET,
 	  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "reset" },
 #endif /* WIRELESS_EXT >= 12 */
